/***************************************************************************
 *   Copyright (C) 2006-2011 by Julien Lemoine, Simon Viennot              *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "crambase.h"

//there is a dependency to gameGlobal because of the use of win, loss, draw in isVictory

#include "../gameglobal.h"
#include <QString>

//--------------------- Move Rules-------------------------

//Indicate whether a move is possible or not
//@return true if the play is valid
//this function doesn't modify the board, contrary to CramBase::play
bool CramBase::possiblePlay(int i, int j, int direction) { 
	if(getCell(i, j) != Empty) return false; //the current cell is not empty
	if( ! next(i, j, direction) ) return false; //invalid direction : we are out of the board
	if(getCell(i, j) != Empty) return false; //the adjacent cell is not empty
	return true;
}

//! Play at the point of coordinates i, j in the given direction.
//! @return true if the play is valid
//! Play is valid iff the two adjacent cells are empty.
bool CramBase::play(int i, int j, int direction) { 
	int i0 = i, j0 = j;
	
	if(getCell(i, j) != Empty) return false; //the current cell is not empty
	if( ! next(i, j, direction) ) return false; //invalid direction : we are out of the board
	if(getCell(i, j) != Empty) return false; //the adjacent cell is not empty
	
	//the two adjacent cells were empty, so the move is valid
	setCell(i0, j0, Gray);
	setCell(i, j, Gray);
	
	return true;
}

//--------------------- Board Reduction-------------------------

//! Indicate if the given point is isolated or not
bool CramBase::isolated(int i0, int j0) {
	int i, j, k;
	//0, 1, 2, 3 are the 4 vertical and horizontal directions
	for(k=0; k<4; k++) {
		i = i0;
		j=j0;
		if(next(i, j, k)) {
			//the adjacent cell in this direction is valid
			//If it is empty, then the given (i0, j0) point is not isolated
			if(getCell(i, j) == Empty) return false;
		}
	}
	
	return true;
}

//TODO : fillIsolatedCells could be optimized because only neighbour cells from the last move can become isolated
//! Fill isolated cells with Gray color
void CramBase::fillIsolatedCells() {
	int i, j;
	for(i=0; i<rows; i++) {
		for(j=0; j<columns; j++) {
			if(isolated(i, j)) {
				setCell(i, j, Gray);
			}
		}
	}
}

//! Reduce the size of the board, by deleting all the surrounding filled rows or columns 
void CramBase::reduceSize() {
	//first step : find the smallest containing rectangle between (i0, j0) -> (i1,j1)
	//during the algorithm, (i0, j0) will only decrease, and (i1, j1) will only increase
	int i0=rows-1, j0=columns-1, i1=0, j1=0;
	int i, j;
	for(i=0; i<rows; i++) {
		for(j=0; j<columns; j++) {
			if(getCell(i, j) == Empty ) {
				//an empty cell gives new conditions on the smallest rectangle
				if(i<i0) i0 = i;
				if(j<j0) j0 = j;
				if(i>i1) i1 = i;
				if(j>j1) j1 = j;
			}
		}
	}
	
	//second step : reduce the size of the board to the rectangle i0, j0, i1, j1
	reduceBoardTo(i0, j0, i1, j1);
}

//----------------- Components : general functions -------------

//return true if the position is composed of multiple boards
bool CramBase::isMultiComponents() {
	return (boardList.size() > (unsigned int) 1);
}

//return the list of components as individual positions
list<Line> CramBase::sumComponents() {
	list<Line> result;
	
	//we need to transform each string in a Line
	int index;
	for(index = 0; index < (int) boardList.size(); index++) {
		//initialize the current board
		initCurrentBoard(index);
		result.push_back(toLine());
	}
	
	return result;
}

//---------------------- Algorithm to separate independent components------------------------------------------

//recursively color one connected component of the board
void CramBase::colorOneComponent(int i0, int j0, int colorNumber) {
	//nothing to do if the cell is already colored
	if(colors[i0*columns+j0]>=0) return;
	
	//color the current cell
	colors[i0*columns+j0]=colorNumber;
	
	//color recursively all the adjacent cells
	int direction;
	for(direction=0; direction<4; direction++) {
		int i=i0, j=j0;
		if(next(i, j, direction)) {
			//this is a valid direction, try to extend the connex component
			colorOneComponent(i, j, colorNumber);
		}
	}
}

//color the connected components of the board, with a different color for each component
//return the number of colors, which is also the number of components
int CramBase::colorComponents() {
	colors.clear();
	
	//color 0 means an occupied square, color -1 means an empty square
	int i, j;
	for(i=0; i<rows; i++) {
		for(j=0; j<columns; j++) {
			if(getCell(i,j) == Empty) {
				colors.push_back(-1);
			} else {
				colors.push_back(0);
			}
		}
	}
	
	int colorNumber = 0;
	
	//color the board until all cell are colored
	for(i=0; i<rows; i++) {
		for(j=0; j<columns; j++) {
			if(colors[i*columns + j]<0) {
				colorNumber++;
				colorOneComponent(i, j, colorNumber);
			}
		}
	}
	
	return colorNumber;
}

//Obtain the board corresponding to a given color (=one connected component)
//Cells of the given color are emptied
//Other cells are colored in Gray
void CramBase::getOneComponent(int colorNumber) {
	//cells of the given color are the new empty cells
	int i, j;
	for(i=0; i<rows; i++) {
		for(j=0; j<columns; j++) {
			if(colors[i*columns + j]==colorNumber) {
				setCell(i, j, Empty);
			} else {
				setCell(i, j, Gray);
			}
		}
	}
}

//---------------------- Canonization -----------------------------------------

//by default, canonize with vertical and horizontal symmetry + rotations (=Cram game)
int CramBase::symmetryForCano() {
	return 4;
}

//canonization of a board list
//separate each board into independent components
void CramBase::canonize() {
	if(boardList.empty()) return;

	vector<string> result;
	
	//we need to separate into components each board of the board list
	int index;
	for(index = 0; index < (int) boardList.size(); index++) {
		//initialize the current board
		initCurrentBoard(index);
		int colorNumber = colorComponents();
		saveBoard(); //getOneComponent changes the board, so we save it
		
		int i;
		for(i=1; i<=colorNumber; i++) {
			getOneComponent(i);
			reduceSize();
			symmetry(symmetryForCano());  //canonize the authorized symetries
			result.push_back(toString());
			
			//caution : we need to go back to the initial board, in order to get correctly the next component
			backToSavedBoard();
		}
	}
	
	//update the boardList with the result list
	boardList = result;
	
	//initialize the current board to the first board of the list
	if(!boardList.empty()) initCurrentBoard(0);
}

//---------------------- Move ordering -----------------------------------------

//return the score of position
int CramBase::surroundingCellsScore() {
	//give the priority to positions with a lot of used cells on the surrounding of the board
	int score=0;
	
	int i, j;
	for(i=0; i<rows; i++) {
		for(j=0; j<columns; j++) {
			if(getCell(i,j) != Empty) {
				//this is a used cell, it gives points
				//score is proportionnal to the distance to the center of the board
				score += (rows/2 - i) * (rows/2 - i) + (columns/2 - j) * (columns/2 - j);
			}
		}
	}
	
	return score;
}

// return the length of the shortest "remaining" part needed to complete a straigth cutting line
int CramBase::minCuttingLine() {
	//look for the minimum horizontal line
	int minHor = columns;
	int lineEmpty;
	int i, j;
	for(i=0; i<rows; i++) {
		lineEmpty=0;
		for(j=0; j<columns; j++) {
			if(getCell(i, j) == Empty) {
				lineEmpty++;
			}
		}
		
		if(lineEmpty < minHor) {
			minHor = lineEmpty;
		}
	}

	//look for the minimum vertical line
	int minVer = rows;
	for(j=0; j<columns; j++) {
		lineEmpty=0;
		for(i=0; i<rows; i++) {
			if(getCell(i, j) == Empty) {
				lineEmpty++;
			}
		}
		
		if(lineEmpty < minVer) {
			minVer = lineEmpty;
		}
	}

	//return the minimum between horizontal and vertical candidates
	if(minHor < minVer) {
		return minHor;
	} else {
		return minVer;
	}
}
