/***************************************************************************
 *   Copyright (C) 2006-2011 by Julien Lemoine, Simon Viennot              *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "glopwidget.h"
#include "widgetpopup.h"
#include "trace.h"
#include "nodestore.h"

#include "basegame.h"
#include "basenode.h"

#include <QtGui>

QString GlopWidget::glopVersionNumber("2.0");

//---------------------ClickTextEdit---------------------

ClickTextEdit::ClickTextEdit( QWidget * parent ) : QTextEdit(parent) {
	//nothing to do
}

//double-click implementation for the children tab
void ClickTextEdit::mouseDoubleClickEvent( QMouseEvent * ) {
    // Emit content of the current line
    QTextCursor cursor = textCursor();
    cursor.movePosition(QTextCursor::StartOfLine, QTextCursor::MoveAnchor);
    cursor.movePosition(QTextCursor::EndOfLine, QTextCursor::KeepAnchor);
    QString content = cursor.selectedText();
	emit lineContent(content);
}

//--------------------------------------------------------------------------
//Interface definition
//--------------------------------------------------------------------------

vector<InterfaceWidget *> GlopWidget::gameWidgets;
vector<InterfaceWidget *> GlopWidget::nodeWidgets;
vector<InterfaceWidget *> GlopWidget::traversalWidgets;
vector<InterfaceWidget *> GlopWidget::commonWidgets;
QTabWidget *GlopWidget::tabNodeWidget;
QTabWidget *GlopWidget::tabWidget;
QComboBox *GlopWidget::box_game_choice;
QComboBox *GlopWidget::box_traversal_choice;

GlopWidget::GlopWidget( QWidget *parent) : QMainWindow( parent ) {
	//icon on the upper left corner and in the taskbar
	setWindowIcon(QIcon(QString(":/images/glop-24x32.png") ));
	
	//window title
	setWindowTitle(QString("GLOP"));
	//game object
	game = new Game;
	
	computeRunning = false;
	pauseEnabled = false;
	noMoreMessage = false;
	
	//timer
	impulse = new QTimer(this);
	connect( this, SIGNAL(isComputing(bool)), this, SLOT(activate_timer(bool)) );
	
	//on each impulse, check result state and refresh tables tracing computation
	connect( impulse, SIGNAL(timeout()), this, SLOT(wait_result()) );
	connect( impulse, SIGNAL(timeout()), this, SLOT(trace_refresh()) );
	
	//error handling (check error on each timer impulse)
	WidgetError *widget_error= new WidgetError(this);
	connect( impulse, SIGNAL(timeout()), widget_error, SLOT(receive_error()) );
	
	//connexion mapping
	readText_Mapper=new QSignalMapper(this);
	connect( readText_Mapper, SIGNAL(mapped(int)), this, SLOT(read_text_file(int)) );
	
	//layout
	QWidget *widget_central = new QWidget( this );
	setCentralWidget(widget_central);
	QVBoxLayout *central_Vlayout = new QVBoxLayout(widget_central);	//vertical layout of main window
	
	//-----upper box-------------------------
	QGridLayout *upper_box = new QGridLayout;	//horizontal layout of top part
	central_Vlayout->addLayout( upper_box );
	
	//glop icon
	QLabel *labelIcon = new QLabel( widget_central );
	QPixmap *pixIcon = new QPixmap(":/images/glop-100x46.png");
	labelIcon->setPixmap(*pixIcon);
	labelIcon->setFixedSize( 100, 46 );
	upper_box->addWidget( labelIcon, 0, 0, 3, 1 );
	
	//upper tabs for the node kinds
	tabNodeWidget = new QTabWidget(widget_central);
	upper_box->addWidget( tabNodeWidget, 0, 1, 3, 1);
	connect( this, SIGNAL(isComputing(bool)), tabNodeWidget, SLOT(setDisabled(bool)) );
	
	int i;
	int n = BaseNode::nodesNumber();
	for(i=0; i<n; i++) {
		//initialize the databases
		BaseNode(i, 0 /*useless value*/).createDataStorage();
		
		Interface paramI = BaseNode(i, 0 /*useless value*/).getParamDef();
		InterfaceWidget * paramW = new InterfaceWidget;
		paramW->init(paramI);
		tabNodeWidget->addTab(paramW, QString(paramI.name.c_str()));
		nodeWidgets.push_back(paramW);
		
		//refresh the values on each impulse : used to refresh the databaseButtons
		connect( impulse, SIGNAL(timeout()), paramW, SLOT(refreshValues()) );
	}
	
	//Start button
	QPushButton *start_button = new QPushButton( "Start", widget_central );
	start_button->setFixedSize(start_button->sizeHint());
	upper_box->addWidget( start_button, 0, 2);
	connect( this, SIGNAL(isComputing(bool)), start_button, SLOT(setDisabled(bool)) );
	connect( start_button, SIGNAL(clicked()), this, SLOT(startComputation()) );
	
	//Pause button
	pause_button = new QPushButton( "Pause", widget_central );
	pause_button->setEnabled(false);
	pause_button->setFixedSize(pause_button->sizeHint());
	upper_box->addWidget( pause_button, 1, 2);
	connect( pause_button, SIGNAL(clicked()), this, SLOT(pause_computing()) );
	connect( this, SIGNAL(isComputing(bool)), pause_button, SLOT(setEnabled(bool)) );
	
	//Stop button
	QPushButton *stop_button = new QPushButton( "Stop", widget_central );
	stop_button->setEnabled(false);
	stop_button->setFixedSize(stop_button->sizeHint());
	upper_box->addWidget( stop_button, 2, 2);
	connect( stop_button, SIGNAL(clicked()), this, SLOT(stop_computing()) );
	connect( this, SIGNAL(isComputing(bool)), stop_button, SLOT(setEnabled(bool)) );
	
	//Label to display the state of NodeStore
	nodeStoreState = new QLabel;
	nodeStoreState->setText(QString("AAAAAAAAAAAAAAAAAAAAAAAAAAA\nA\nA\nA\n")); //to set the size
	nodeStoreState->setFixedSize(nodeStoreState->sizeHint());
	nodeStore_refresh();
	upper_box->addWidget( nodeStoreState, 0, 3, 3, 1);
	connect( impulse, SIGNAL(timeout()), this, SLOT(nodeStore_refresh()) );
	
	//-----Tabs in the main window-----------
	tabWidget = new QTabWidget;
	central_Vlayout->addWidget( tabWidget );
	setup_parameters_tab();
	setup_compute_tab();
	setup_test_tab();
//	setup_analysis_tab();
	setup_info_tab();
}

//Parameters tab
void GlopWidget::setup_parameters_tab() {
  //Main widget, layout and connection of this tab
	QWidget *parameters_tab = new QWidget;
	//Tab disabled when computing
	connect( this, SIGNAL(isComputing(bool)), parameters_tab, SLOT(setDisabled(bool)) );
	//Vertical Layout
	QVBoxLayout *parameters_tab_Vlayout = new QVBoxLayout( parameters_tab );
	
	//------------- Game choice --------------------------
	box_game_choice = new QComboBox;
	parameters_tab_Vlayout->addWidget( box_game_choice );
	
	//create a new choice for each registered game
	int nbGames = BaseGame::gamesNumber();
	int i;
	int SproutsIndex=0;	//determine Sprouts' index, because we select it as the default game later
	for(i=0; i<nbGames; i++) {
		Interface paramI = BaseGame(i).getParamDef();
		if(paramI.name==string("Sprouts")) SproutsIndex=i;
		box_game_choice->insertItem(i, QString(paramI.name.c_str()));
		InterfaceWidget * paramW = new InterfaceWidget;
		paramW->init(paramI);
		paramW->setVisible(i==0);
		parameters_tab_Vlayout->addWidget(paramW);
		gameWidgets.push_back(paramW);
	}
	
	box_game_choice->setFixedSize(box_game_choice->sizeHint());
	connect( box_game_choice, SIGNAL(currentIndexChanged(int)), this, SLOT(gameChoice(int)) );
	box_game_choice->setCurrentIndex(SproutsIndex); //choice by default

	//------------- Traversal choice --------------------------
	box_traversal_choice = new QComboBox;
	parameters_tab_Vlayout->addWidget( box_traversal_choice );
	
	//create a new choice for each registered traversal algorithm
	int nbTraversal = BaseTraversal::classesNumber();
	for(i=0; i<nbTraversal; i++) {
		Interface paramI = BaseTraversal(i).getParamDef();
		box_traversal_choice->insertItem(i, QString(paramI.name.c_str()));
		InterfaceWidget * paramW = new InterfaceWidget;
		paramW->init(paramI);
		paramW->setVisible(i==0 && nbTraversal>1);
		parameters_tab_Vlayout->addWidget(paramW);
		traversalWidgets.push_back(paramW);
	}
	
	box_traversal_choice->setFixedSize(box_traversal_choice->sizeHint());
	connect( box_traversal_choice, SIGNAL(currentIndexChanged(int)), this, SLOT(traversalChoice(int)) );
	box_traversal_choice->setCurrentIndex(0);
	if(nbTraversal<=1) box_traversal_choice->setVisible(false);
	
	//--------- common interfaces ------------------------------------
	vector<Interface> commonInterfaces;
	
	//Options
	Interface interface2;
	interface2.title=string("Options");
	interface2.addLabel(string("Refresh time (x 100 ms) :"), /*pos*/ 0, 0);
	interface2.addSpinBox(1, 10, 1, /*pos*/ 0, 1);
	interface2.link(refreshTime);
	//interface2.addCheckButton(string("Use graph :"), false, /*pos*/ 1, 0);
	//interface2.link(game->check_graph.use_graph);
	interface2.addLabel(string("Debug level :"), /*pos*/ 0, 2);
	interface2.addSpinBox(0, 10, 0, /*pos*/ 0, 3);
	interface2.link(NodeStore::debugLevel);
	interface2.addCheckButton(string("Use the minimal recursive structure (no trace)"), false, 2, 0);
	interface2.link(game->recursiveStructure);	
	commonInterfaces.push_back(interface2);
	
	//---- Traversal ---
	//Interface traversalInterface = PN_Search::getParamDef();
	//commonInterfaces.push_back(traversalInterface);
	
	//Graph options
	//Graph options are not available until a new Graph object is written
	/*
	Interface interface3;
	interface3.title=string("Graph Options");
	interface3.addCheckButton(string("Use colors :"), false, 0, 0);
	interface3.link(game->check_graph.use_colors);
	interface3.addLabel(string("Detail level :"), 1, 0);
	interface3.addSpinBox(1, 3, 2, 1, 1);
	interface3.link(game->check_graph.detail_level);
	interface3.addLabel(string("Minimal number of lives :"), 2, 0);
	interface3.addSpinBox(0, 99, 0, 2, 1);
	interface3.link(game->check_graph.min_lives);
	commonInterfaces.push_back(interface3);
	*/
	
	//add the common interfaces to glopwidget
	vector<Interface>::iterator Li;
	for(Li=commonInterfaces.begin(); Li!=commonInterfaces.end(); Li++) {
		InterfaceWidget * paramW = new InterfaceWidget;
		paramW->init(*Li);
		parameters_tab_Vlayout->addWidget(paramW);
		commonWidgets.push_back(paramW);
	}
		
	//add Parameter tab to the tabWidget
	tabWidget->addTab(parameters_tab, "Parameters");
}

void GlopWidget::gameChoice(int index) {
	Parameter::gameChoice = index;
	int i;
	int nbGames = BaseGame::gamesNumber();
	for(i=0; i<nbGames; i++) {
		InterfaceWidget * paramW = gameWidgets[i];
		paramW->setVisible(i==index);
	}
	
	//display only the possible nodes for the game category
	selectPossibleNode(BaseGame(Parameter::gameChoice).gameCategory());
}

void GlopWidget::traversalChoice(int index) {
	Parameter::traversalChoice = index;
	int i;
	int nbClasses = BaseTraversal::classesNumber();
	for(i=0; i<nbClasses; i++)  {
		InterfaceWidget * paramW = traversalWidgets[i];
		paramW->setVisible(i==index && nbClasses>1);
	}
}

void GlopWidget::selectPossibleNode(int gameCategory) {
	int i;
	int n = BaseNode::nodesNumber();
	bool currentIndexCorrect=false;
	int firstCorrectIndex=n-1;
	int nimberIndex=-1;	//We want NimberNode as the default Node for Impartial games
	
	for(i=0; i<n; i++) {
		GameType t = BaseNode(i, 0 /*useless value*/).gameType();
		
		Interface paramI = BaseNode(i, 0).getParamDef();//these 2 lines allow us to find the Nimber Index
		if(paramI.name==string("Nimber")) nimberIndex=i;
		
		bool possibleNode = t.isFor[gameCategory];
		tabNodeWidget->setTabEnabled(i, possibleNode);
		if(tabNodeWidget->currentIndex()==i) {
			currentIndexCorrect = possibleNode;
		}
		if(possibleNode && i< firstCorrectIndex) {
			firstCorrectIndex = i;
		}
	}
	
	//For an Impartial game, select "Nimber" tab
	if(gameCategory==GameType::ImpartialGame) {
		tabNodeWidget->setCurrentIndex(nimberIndex);
	} else
	//set the current tab to a valid tab (if needed)
	if(!currentIndexCorrect) {
		tabNodeWidget->setCurrentIndex(firstCorrectIndex);
	}
}

//Computation and Explore tab
void GlopWidget::setup_compute_tab() {
  //Layout for this tab
	QWidget *compute_tab = new QWidget;			//main widget of Computation tab
	//connect( this, SIGNAL(isChecking(bool)), compute_tab, SLOT(setDisabled(bool)) );
	QVBoxLayout *compute_tab_Vlayout = new QVBoxLayout( compute_tab );	//layout for this tab
	compute_tab_Vlayout->setSpacing(8);
	compute_tab_Vlayout->setMargin(5);
	
  //Table to trace computation
	widget_trace=new WidgetTrace(compute_tab);
	connect( this, SIGNAL(isComputing(bool)), widget_trace->table, SLOT(setEnabled(bool)) );
	connect( widget_trace, SIGNAL(requestExplorer()), this, SLOT(displayExplorer()) );
	compute_tab_Vlayout->addWidget( widget_trace );
	
	//Explore tab
	explore_tab = new QWidget;			//main widget of Explore tab
	//connect( this, SIGNAL(isChecking(bool)), compute_tab, SLOT(setDisabled(bool)) );
	QVBoxLayout *explore_tab_Vlayout = new QVBoxLayout( explore_tab );	//layout for this tab
	explore_tab_Vlayout->setSpacing(8);
	explore_tab_Vlayout->setMargin(5);
	widget_trace->children_popup->requestedId = 0;
	connect( this, SIGNAL(isComputing(bool)), widget_trace->children_popup, SLOT(setEnabled(bool)) );
	explore_tab_Vlayout->addWidget( widget_trace->children_popup );
	
  //add Computation tab to tabWidget
	tabWidget->addTab( compute_tab, "Computing branch" );
	tabWidget->addTab( explore_tab, "Search tree" );
}

//display the explorer (called when we click on a position of the computation tab)
void GlopWidget::displayExplorer() {
	tabWidget->setCurrentWidget(explore_tab);
}

/*//Analysis tab
void GlopWidget::setup_analysis_tab() {
	QWidget *analysis_tab = new QWidget;
	connect( this, SIGNAL(isComputing(bool)), analysis_tab, SLOT(setDisabled(bool)) );
	QVBoxLayout *analysis_tab_Vlayout = new QVBoxLayout( analysis_tab );
	QLabel *filterTitle = new QLabel( analysis_tab );
	filterTitle->setText( "Filter database with the following parameters :" );
	filterTitle->setFixedSize(filterTitle->sizeHint());
	analysis_tab_Vlayout->addWidget( filterTitle );
	
  //Number of lives
	QGroupBox *livesGroup = new QGroupBox("Number of lives", analysis_tab);
	analysis_tab_Vlayout->addWidget( livesGroup );
	QGridLayout *lives_grid = new QGridLayout( livesGroup );
	
	//Minimum number of lives
	check_lives_min = new QCheckBox( livesGroup );
	lives_grid->addWidget( check_lives_min, 0, 0 );
	
	QLabel *lives_min_label = new QLabel( livesGroup );
	lives_min_label->setText("Minimum number of lives :");
	lives_min_label->setEnabled(false);
	connect( check_lives_min, SIGNAL(toggled(bool)), lives_min_label, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( lives_min_label, 0, 1 );
	
	box_lives_min = new QSpinBox( livesGroup );
	box_lives_min->setMinimum(0);
	box_lives_min->setMaximum(99);
	box_lives_min->setFixedSize(box_lives_min->sizeHint());
	box_lives_min->setEnabled(false);
	connect( check_lives_min, SIGNAL(toggled(bool)), box_lives_min, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( box_lives_min, 0, 2 );
	
	//Maximum number of lives
	check_lives_max = new QCheckBox( livesGroup );
	lives_grid->addWidget( check_lives_max, 1, 0 );
	
	QLabel *lives_max_label = new QLabel( livesGroup );
	lives_max_label->setText("Maximum number of lives :");
	lives_max_label->setEnabled(false);
	connect( check_lives_max, SIGNAL(toggled(bool)), lives_max_label, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( lives_max_label, 1, 1 );
	
	box_lives_max = new QSpinBox( livesGroup );
	box_lives_max->setMinimum(0);
	box_lives_max->setMaximum(99);
	box_lives_max->setFixedSize(box_lives_max->sizeHint());
	box_lives_max->setEnabled(false);
	connect( check_lives_max, SIGNAL(toggled(bool)), box_lives_max, SLOT(setEnabled(bool)) );
	lives_grid->addWidget( box_lives_max, 1, 2 );
	
  //Nimber
	QGroupBox *nimberGroup = new QGroupBox("Nimber", analysis_tab);
	analysis_tab_Vlayout->addWidget( nimberGroup );
	QGridLayout *nimber_grid = new QGridLayout( nimberGroup );
	
	//Minimum nimber
	check_nimber_min = new QCheckBox( nimberGroup );
	nimber_grid->addWidget( check_nimber_min, 0, 0 );
	
	QLabel *label_nimber_min = new QLabel( nimberGroup );
	label_nimber_min->setText("Minimum nimber :");
	label_nimber_min->setEnabled(false);
	connect( check_nimber_min, SIGNAL(toggled(bool)), label_nimber_min, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( label_nimber_min, 0, 1 );
	
	box_nimber_min = new QSpinBox( nimberGroup );
	box_nimber_min->setMinimum(0);
	box_nimber_min->setMaximum(99);
	box_nimber_min->setFixedSize(box_nimber_min->sizeHint());
	box_nimber_min->setEnabled(false);
	connect( check_nimber_min, SIGNAL(toggled(bool)), box_nimber_min, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( box_nimber_min, 0, 2 );
	
	//Maximum nimber
	check_nimber_max = new QCheckBox( nimberGroup );
	nimber_grid->addWidget( check_nimber_max, 1, 0 );
	
	QLabel *label_nimber_max = new QLabel( nimberGroup );
	label_nimber_max->setText("Maximum nimber :");
	label_nimber_max->setEnabled(false);
	connect( check_nimber_max, SIGNAL(toggled(bool)), label_nimber_max, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( label_nimber_max, 1, 1 );
	
	box_nimber_max = new QSpinBox( nimberGroup );
	box_nimber_max->setMinimum(0);
	box_nimber_max->setMaximum(99);
	box_nimber_max->setFixedSize(box_nimber_max->sizeHint());
	box_nimber_max->setEnabled(false);
	connect( check_nimber_max, SIGNAL(toggled(bool)), box_nimber_max, SLOT(setEnabled(bool)) );
	nimber_grid->addWidget( box_nimber_max, 1, 2 );
	
  //Buttons
	QHBoxLayout *buttons_Hlayout = new QHBoxLayout;
	analysis_tab_Vlayout->addLayout( buttons_Hlayout );
	
	QPushButton *start_button = new QPushButton( "Start", analysis_tab);
	start_button->setFixedSize(start_button->sizeHint());
	connect( start_button, SIGNAL(clicked()), startCompute_Mapper, SLOT(map()) );
	startCompute_Mapper->setMapping(start_button, global::Analysis);
	buttons_Hlayout->addWidget( start_button );
	
	//WARNING : analysis is temporarily disabled until class Game is separated in different modules	
	start_button->setEnabled(false);

	tabWidget->addTab( analysis_tab, "Analysis" );
}*/

//Test tab
void GlopWidget::setup_test_tab() {
	QWidget *test_tab = new QWidget;
	connect( this, SIGNAL(isComputing(bool)), test_tab, SLOT(setDisabled(bool)) );
	QVBoxLayout *test_tab_Vlayout = new QVBoxLayout( test_tab );
  //Children computation
	QGroupBox *children_group = new QGroupBox("Children computation", test_tab);
	test_tab_Vlayout->addWidget( children_group );
	QGridLayout *grid_test = new QGridLayout( children_group );
	
	QLabel *repr_2 = new QLabel( children_group );
	repr_2->setText("Position :");
	grid_test->addWidget( repr_2, 0, 0 );
	box_given_position_2 = new QLineEdit( children_group );
	grid_test->addWidget( box_given_position_2, 0, 1 );
	box_given_position_2->setText( QString("0*4.AB|0*4.AB") );
	box_given_position_2->setMinimumWidth( 150 );
	
	QPushButton *back_button = new QPushButton( "Back to previous", children_group );
	back_button->setFixedSize(back_button->sizeHint());
	grid_test->addWidget( back_button, 1, 0 );
	connect( back_button, SIGNAL(clicked()), this, SLOT(previousChildren()) );
	
	QPushButton *children_button = new QPushButton( "Compute children", children_group );
	children_button->setFixedSize(children_button->sizeHint());
	grid_test->addWidget( children_button, 2, 0 );
	connect( children_button, SIGNAL(clicked()), this, SLOT(printChildren()) );
	
	textEdit_children = new ClickTextEdit( children_group );
	textEdit_children->setReadOnly( TRUE );
	connect(textEdit_children ,SIGNAL(lineContent(QString)), this, SLOT(nextChildren(QString)));
	grid_test->addWidget( textEdit_children, 1, 1, 2, 1);
	
	tabWidget->addTab( test_tab, "Children" );
}

//Info tab
void GlopWidget::setup_info_tab() {
	QWidget *info_tab = new QWidget;
	QGridLayout *grid_info = new QGridLayout( info_tab );
	
  //Images
	QLabel *labelHelp = new QLabel( info_tab );
	QPixmap *pixHelp = new QPixmap(":/images/help64.png");
	labelHelp->setPixmap(*pixHelp);
	labelHelp->setFixedSize( 64, 64 );
	grid_info->addWidget( labelHelp, 0, 0 );
	
	QLabel *labelLicence = new QLabel( info_tab );
	QPixmap *pixLicence = new QPixmap(":/images/licence64.png");
	labelLicence->setPixmap(*pixLicence);
	labelLicence->setFixedSize( 64, 64 );
	grid_info->addWidget( labelLicence, 0, 1 );
	
	QLabel *labelAbout = new QLabel( info_tab );
	QPixmap *pixAbout = new QPixmap(":/images/about64.png");
	labelAbout->setPixmap(*pixAbout);
	labelAbout->setFixedSize( 64, 64 );
	grid_info->addWidget( labelAbout, 0, 2 );
	
  //Buttons
	QPushButton *help_button = new QPushButton( "Help", info_tab );
	help_button->setFixedSize(help_button->sizeHint());
	grid_info->addWidget( help_button, 1, 0 );
	connect( help_button, SIGNAL(clicked()), readText_Mapper, SLOT( map() ) );
	readText_Mapper->setMapping(help_button, 0);
	
	QPushButton *licence_button = new QPushButton( "Licence", info_tab );
	licence_button->setFixedSize(licence_button->sizeHint());
	grid_info->addWidget( licence_button, 1, 1 );
	connect( licence_button, SIGNAL(clicked()), readText_Mapper, SLOT( map() ) );
	readText_Mapper->setMapping(licence_button, 1);
	
	QPushButton *about_button = new QPushButton( "About", info_tab );
	about_button->setFixedSize(about_button->sizeHint());
	grid_info->addWidget( about_button, 1, 2 );
	connect( about_button, SIGNAL(clicked()), readText_Mapper, SLOT( map() ) );
	readText_Mapper->setMapping(about_button, 2);
	
	tabWidget->addTab( info_tab, "Information" );
}


//--------------------------------------------------------------------------
//Starts/Stops computation
//--------------------------------------------------------------------------

void GlopWidget::startComputation() {
	int n = tabNodeWidget->currentIndex();
	Parameter::nodeChoice = n;
	
	//read the parameters of this kind of node
	if(n>=0 && n< (int) nodeWidgets.size()) {
		(nodeWidgets[n])->copyInTarget();
	} else {
		//invalid tab
		return;
	}
	
	Parameter::computation_kind = BaseNode(n, 0 /*useless value*/).startComputationKind();
	
	Parameter::debugGames=false; //initialize the debug variable for games (no debug)
	
	//Read user-controled Widget parameters
//	if(computation_kind==global::Analysis) {
//		read_analysis_parameters(); //Analysis paramaters
//	} 

	//read the parameters specific to the chosen game
	InterfaceWidget * paramW = gameWidgets[Parameter::gameChoice];
	paramW->copyInTarget();
	BaseGame(Parameter::gameChoice).setParam();
	
	//read the parameters of the chosen traversal
	traversalWidgets[Parameter::traversalChoice]->copyInTarget();
	
	//read the parameters of all the common widgets
	vector<InterfaceWidget *>::iterator Wi;
	for(Wi=commonWidgets.begin(); Wi!=commonWidgets.end(); Wi++) {
		(*Wi)->copyInTarget();
	}
	
  //initialize the trace
	Trace::initialize();
	
  //signals that computation started
	Trace::set_stop_value(false);
	game->computationRunning=true;
	computeRunning=true;
	emit isComputing(TRUE);
	
	//call Game::run() as a separate thread with the special start() function
	game->start(QThread::NormalPriority);
}

//Read user-defined analysis parameters
/*void GlopWidget::read_analysis_parameters() {
  //number of lives
	game->lives_min_analysis=check_lives_min->isChecked();
	if(game->lives_min_analysis) game->lives_min=box_lives_min->value();
	game->lives_max_analysis=check_lives_max->isChecked();
	if(game->lives_max_analysis) game->lives_max=box_lives_max->value();
  //nimber
	game->nimber_min_analysis=check_nimber_min->isChecked();
	if(game->nimber_min_analysis) game->nimber_min=box_nimber_min->value();
	game->nimber_max_analysis=check_nimber_max->isChecked();
	if(game->nimber_max_analysis) game->nimber_max=box_nimber_max->value();
}*/

//stops computation
void GlopWidget::stop_computing() {
	//launch the signal of end of computation
	Trace::set_stop_value(true);
	
	//end the pause if necessary
	if(pauseEnabled) {
		Trace::pauseEnd();
		pause_button->setText("Pause");
		pauseEnabled = false;
	}
}

void GlopWidget::pause_computing() {
	if(pauseEnabled) {
		Trace::pauseEnd();
		pause_button->setText("Pause");
	} else {
		Trace::pauseStart();
		pause_button->setText("Restart");
	}
	pauseEnabled = !pauseEnabled;
}

//--------------------------------------------------------------------------
//Test functions
//--------------------------------------------------------------------------

//prints children of a given position
void GlopWidget::printChildren() {
	textEdit_children->setPlainText("");	//clean text box
	
	//add the position to the cache (used to go back to a previous position)
	positionCache.push_back(box_given_position_2->text()); 
	
	//currently selected nodeChoice
	int n = tabNodeWidget->currentIndex();
	Parameter::nodeChoice = n;
	
	//init the possible internal variables of this kind of node
	//note : it is usually used to initialize the first position of a computation
	//but it is also used by some nodes to initialize static variables
	BaseNode tmpNode(Parameter::nodeChoice, Parameter::computation_kind);
	tmpNode.initRootNode();
	
	//compute the children
	BaseGame g((box_given_position_2->text()).toStdString());
	g.computeChildrenSet();
	
	set<Line>::iterator it;
	list<Line> result;
	for(it=g.childrenSet().begin(); it!=g.childrenSet().end(); it++) {
		BaseGame h(*it);
		h.canonize();
		result.push_back(h.exportToLine());
	}
	
	result.sort();		//needed for unique
	result.unique();	//removes all consecutive duplicate positions
	
	//print children in the text box (use zero short notation)
	list<Line>::iterator Pi;
	for(Pi=result.begin();Pi!=result.end();Pi++){
		BaseGame h(*Pi);
		textEdit_children->insertPlainText(QString(h.exportToString().c_str()));
		
		//print info about the current type of node
		BaseNode c(Parameter::nodeChoice, 0 /*useless value*/);
		QString info = c.infoAbout(h.exportToString());
		textEdit_children->insertPlainText(QString(" ") + info + QString("\n"));
	}
}

//compute the children of a given string
//called after a double-click in the children tab
void GlopWidget::nextChildren(QString position) {
	//keep only the position part (without nimber)
	int a = position.indexOf(' ');
	QString positionPart = position.left(a);
	
	//compute children
	box_given_position_2->setText(positionPart);
	printChildren();
}

//Go back to the previous position
void GlopWidget::previousChildren() {
	if(positionCache.size() >= 2){
		positionCache.pop_back(); //delete the current child
		QString previous = positionCache.back();
		positionCache.pop_back();
		box_given_position_2->setText(previous);
		printChildren();
	}
}

//--------------------------------------------------------------------------
//Functions to trace computation
//--------------------------------------------------------------------------

//timer is activated when we receive true
//activate_timer is connected to isComputing
void GlopWidget::activate_timer(bool val) {
	if(val) {
		impulse->start(refreshTime*100);
	} else {
		impulse->stop();
	}
}

//refresh widgets on each timer's impulse
void GlopWidget::trace_refresh() {
	widget_trace->update_display();
}

//refresh the label with the state of NodeStore on each timer's impulse
void GlopWidget::nodeStore_refresh() {
	nodeStoreState->setText(NodeStore::stateString());
}

//Print Computation result
void GlopWidget::wait_result() {
	bool isRunning;
	
	if(!game->game_lock.tryLock()) return;		//return if lock failed
		isRunning=game->computationRunning;
	game->game_lock.unlock();
	
	if (isRunning) return; //computation is still running
	
	emit isComputing(FALSE); //signals computation ended
	emit isChecking(FALSE);
	computeRunning=false;
	
	if(noMoreMessage) {
		//user has asked to quit the program as soon as computation ends
		qApp->exit(0);
		return;
	}
	
	QString positionName;
	positionName = BaseGame(Parameter::gameChoice).getPositionName().c_str();
	
	//display message
	QString message;
	QString message_title="Result";
	/*if (computation_kind==global::Analysis) {
		message+="Analysis is completed.";
	}*/
	/*
	if (computation_kind==global::Complete_check) {
		if (result==global::Loss) {
			message+="Complete check stopped.";
		} else {
			message+="Database successfully checked";
		}
	}
	*/
	
	if(Trace::stop_signal()) {
		message+="Computation stopped.";
	} else {
		message += QString("Game     : %1\n").arg(box_game_choice->currentText());
		message += QString("Position : %1\n").arg(positionName);
		message += QString("Result    : %1\n").arg(BaseNode::computationResult.c_str());
	}
	
	QMessageBox::information (this, message_title, message, QMessageBox::Ok);
}

//open the right popup
void GlopWidget::read_text_file(int num) {
	if(num==0) {
		WidgetPopup popup( this, QString(":/txt/help.html"), QString("Help"));
		popup.exec();
	} else if(num==1) {
		WidgetPopup popup( this, QString(":/txt/gpl.html"), QString("Licence"));
		popup.exec();
	} else if(num==2) {
		WidgetPopup popup( this, QString(":/txt/about.html"), QString("About"));
		popup.exec();
	}
}

//check that databases are saved when user closes the program
void GlopWidget::closeEvent(QCloseEvent *event) {
	if( !computeRunning && db::allSaved()) {
		//accept the close event (end of the program)
		event->accept();
		return;
	}
	
	//determine error message
	QString message;
	if(computeRunning) {
		message += "A computation is running.";
	} else {
		message += db::unsavedBaseName() + " base has not been saved.";
	}
	
	message += "\nDo you really want to quit?";
	
	//print message
	int cancel=QMessageBox::question(this, "Confirmation", message, "Yes", "No", QString(), 1, 1);
	
	if(cancel) {
		//user canceled : ignore the close event
		event->ignore();
	} else {
		if(computeRunning) {
			//user chose to quit : stop the computation with noMoreMessage set to true
			noMoreMessage = true;
			stop_computing();
			event->ignore(); //the program will quit as soon as computation ends
		} else {
			event->accept();
		}
	}
}
