/***************************************************************************
 *   Copyright (C) 2006-2011 by Julien Lemoine, Simon Viennot              *
 *   lapinot@tuxfamily.org                                                 *
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 *   This program is distributed in the hope that it will be useful,       *
 *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
 *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
 *   GNU General Public License for more details.                          *
 *                                                                         *
 *   You should have received a copy of the GNU General Public License     *
 *   along with this program; if not, write to the                         *
 *   Free Software Foundation, Inc.,                                       *
 *   59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.             *
 ***************************************************************************/

#include "nodestore.h"
#include "trace.h"

#include <QMutex>

void printList(list<storeId> Lid) {
	list<storeId>::const_iterator Si;
	for(Si=Lid.begin(); Si!=Lid.end(); Si++) {
		cout << "#" << *Si << " ";
	}
	cout << endl;
}

NodeData::NodeData() {
	knownState=false;
	studied_pos_index=-1;
	children_number=0;
	unknown_number=0;
	
	//init traversal_data with the correct traversal object
	BaseTraversal bt(Parameter::traversalChoice);
	traversal_data = bt;
	traversal_data.init();
}

//-----------static objects of NodeStore ------------------
QMutex NodeStore::nodeStoreLock;
int NodeStore::debugLevel=0;
int NodeStore::mainloopCount=0;
int NodeStore::TRnodesCount=0;  //count the number of TRaversal nodes
int NodeStore::DFnodesCount=0;  //count the number of Depth-First nodes

storeId NodeStore::nextId=0;
storeId NodeStore::rootId=0;
map<string, storeId> NodeStore::idMap; //id corresponding to the string representation of a BaseNode
map<storeId, BaseNode> NodeStore::nodeMap; //BaseNode corresponding to a given id
map<storeId, NodeData> NodeStore::nodeDataMap; //NodeData corresponding to a given id

//list of children and list of parents
map< storeId, list<storeId> > NodeStore::childrenLink;
map< storeId, list<storeId> > NodeStore::parentLink;

list<storeId> NodeStore::computingBranch;
storeId NodeStore::lastComputedId = -1;

bool NodeStore::useTranspositions=false;

//static variable indicating if we are using a tree traversal algorithm (different from alpha-beta)
bool NodeStore::useTraversal=false; 
storeId NodeStore::userChosenId=-1;

//enable/disable warning messages
//in most cases, warning messages are better
//a notable execption is the children trace because we can try to accept to nodes that don't exist
bool NodeStore::warningEnabled=true; 

BaseTraversal NodeStore::traversalStatic; //object to call "static" functions of BaseTraversal

//--------------implementation of NodeStore ----------------

NodeStore::NodeStore() {
	nextId = 0;
	lastComputedId = 1;
}

storeId NodeStore::getRootId() {
	QMutexLocker locker(&nodeStoreLock);
	return rootId;
}

void NodeStore::clearNodes() {
	QMutexLocker locker(&nodeStoreLock);
	idMap.clear();
	nodeMap.clear();
	nodeDataMap.clear();
	childrenLink.clear();
	parentLink.clear();
	nextId = 0;
	DFnodesCount=0;
	TRnodesCount=0;
	computingBranch.clear();
	lastComputedId = 1;
}

storeId NodeStore::addNode(/*const*/ BaseNode& a) {
	QMutexLocker locker(&nodeStoreLock);
	if(!a.validPointer()) {
		cout << "NodeStore::addNode : Invalid pointer of a BaseNode in NodeStore::addNode" << endl;
	}
	
	string nodeString = a.nodeStoreString();
	
	//transpositions : don't create a new node, if the same node already exists
	if(NodeStore::useTranspositions) {
		map<string, storeId>::iterator it = idMap.find(nodeString);
		if(it!=idMap.end()) {
			if(debugLevel==4) {
				cout << "(addNode) Exist id #" << it->second << "  " << nodeString << endl;
			}
			return it->second;
		}
	}
	
	//add the node with a new storeId
	storeId result = nextId;
	//idMap[nodeString]=nextId;
	nodeMap[nextId]=a;
	if(debugLevel==4) {
		cout << "(addNode) New id #" << nextId << "  " << nodeString << endl;
	}
	nextId++;
	if(nextId<0) cout << "Too long computation. Node Id overflow in NodeStore." << endl;
	
	DFnodesCount++;
	
	//when we store a new node, add also a "void" nodeData
	NodeData d;
	if(useTraversal) d.traversal_data.initFromNode(a);		//Warning : initFromNode is authorized to modify the object a
	nodeDataMap[result] = d;
	return result;
}

//the lock is done in addNode
list<storeId> NodeStore::addNodeList(/*const*/ list<BaseNode>& LNC) {
	list<storeId> result;
	list<BaseNode>::iterator NCi;
	for(NCi=LNC.begin(); NCi!=LNC.end(); NCi++) {
		result.push_back(addNode(*NCi));
	}
	return result;
}

void NodeStore::updNode(const storeId &id, /*const*/ BaseNode& a) {
	QMutexLocker locker(&nodeStoreLock);
	if(!a.validPointer()) {
		cout << "Invalid pointer of BaseNode in NodeStore::updNode #" << id << " " << endl;;
	}
	
	//first step : find the current node for this id
	map<storeId, BaseNode>::iterator it = nodeMap.find(id);
	if(it!=nodeMap.end()) {
		if(NodeStore::useTranspositions) {
			//get the string representation of the old node
			string s = it->second.nodeStoreString();
			
			//delete the string from the index
			deleteNodeString(s);
			
			//update the string index
			idMap[a.nodeStoreString()]=id;
		}
		
		//update the nodeMap
		it->second = a;
	} else {
		cout << "NodeStore::updNode error : id doesn't exist" << endl;
		return;
	}
}

//add links parent->children and children->parent
void NodeStore::addLink(storeId parentId, list<storeId> childrenIdList) {
	QMutexLocker locker(&nodeStoreLock);
	//cout << "Parent id: " << parentId << " childrenId: ";
	//printList(childrenIdList);
	map< storeId, list<storeId> >::iterator it = childrenLink.find(parentId);
	if(it!=childrenLink.end()) {
		//cout << "dead code addLink ? " << endl;
		//cout << "Old link: ";
		list<storeId> tmp=it->second;
		//printList(tmp);
		tmp.insert(tmp.end(), childrenIdList.begin(), childrenIdList.end());
		it->second=tmp;
	} else {
		//this parenId has no children yet
		childrenLink[parentId]=childrenIdList;
		if(debugLevel==4) {
			cout << "(addlink) parent #" << parentId << " children ";
			printList(childrenIdList);
		}
	}
	
	//look for the parent of each children to add parentId
	list<storeId>::const_iterator childId;
	for(childId=childrenIdList.begin(); childId!=childrenIdList.end(); childId++) {
		it = parentLink.find(*childId);
		if(it!=parentLink.end()) {
			list<storeId> parentList = it->second;
			parentList.push_back(parentId);
			it->second = parentList;
		} else {
			//this is the first parent of this node
			list<storeId> parentList;
			parentList.push_back(parentId);
			parentLink[*childId] = parentList;
		}
	}
}

//delete the children link of the node
//indicate in the node data that there are no more children
void NodeStore::deleteChildrenLink(const storeId &id) {
	QMutexLocker locker(&nodeStoreLock);
	
	//------------indicate that the node has no more children------
	bool correctId; //not used
	NodeData data = getNodeData(id, correctId);  //nodeData
	data.children_number=0;
	data.studied_pos_index=-1;
	nodeDataMap[id] = data;
	
	//---------delete the links to the children (recursive)--------
	internalDeleteChildrenLink(id);
}

//internal function (no lock)
void NodeStore::internalDeleteChildrenLink(const storeId &id) {
	if(id<0 || id>nextId) return; //id is out of the posible range
	//cout << "D children of " << id << endl;
	
	//find children and recursively erase them
	map< storeId, list<storeId> >::iterator it2 = childrenLink.find(id);
	if(it2==childrenLink.end()) return;
	
	list<storeId> Lid = it2->second;
	list<storeId>::const_iterator Si;
	for(Si=Lid.begin(); Si!=Lid.end(); Si++) {
		deleteParentLink(id, *Si);
	}
	
	childrenLink.erase(it2);
}

//internal function (no lock)
void NodeStore::deleteParentLink(const storeId &parentId, const storeId &id) {
	if(debugLevel==4) {
		cout << "(DeleteParentLink) Parent : #" << parentId << " Child : #" << id << " " << endl;
	}
	
	//erase the parent from the list of parents
	map< storeId, list<storeId> >::iterator it = parentLink.find(id);
	
	if(it!=parentLink.end()) {
		list<storeId> parentList = it->second;
		parentList.remove(parentId);
		
		//if there is no more parent for the node, delete the node
		if(parentList.empty()) {
			//delete the parentLink
			parentLink.erase(it);
			
			//delete the node
			deleteNode(id);
		} else {
			it->second = parentList;
		}
	} else {
		//there is no parent, delete the node
		deleteNode(id);
	}
}

//internal function (no lock)
//a node delete recursively all its children
void NodeStore::deleteNode(const storeId &id) {
	if(id<0 || id>nextId) return; //id is out of the posible range
	if(debugLevel==4) {
		cout << "(DeleteNode) #" << id << " " << endl;
	}
	
	internalDeleteChildrenLink(id);
	
	//cout << "D id : " << id << endl;
	map<storeId, BaseNode>::iterator it = nodeMap.find(id);
	if(it!=nodeMap.end()) {
		if(NodeStore::useTranspositions) {
			//erase the string representation of the node from the index
			deleteNodeString(it->second.nodeStoreString());
		}
		//erase the node from nodeMap
		nodeMap.erase(it);
	}
	
	deleteNodeData(id);
}

//internal function (no lock)
void NodeStore::deleteNodeData(const storeId &id) {
	map<storeId, NodeData>::iterator it = nodeDataMap.find(id);
	if(it!=nodeDataMap.end()) {
		//we are deleting a node : update the DF or traversal number of nodes
		if(useTraversal && it->second.traversal_data.isTraversalNode()) {
			TRnodesCount--;
		} else {
			DFnodesCount--;
		}
		
		nodeDataMap.erase(it);
	}
}

//internal function (no lock)
//delete a string from the <string, id> map
void NodeStore::deleteNodeString(const string s) {
	map<string, storeId>::iterator it = idMap.find(s);
	if(it!=idMap.end()) {
		idMap.erase(it);
	}
}

//external function : call the internal function with a lock
BaseNode NodeStore::externGetBaseNode(const storeId &id, bool &correctId) {
	QMutexLocker locker(&nodeStoreLock);
	return getBaseNode(id, correctId);
}

//external function : call the internal function with a lock
NodeData NodeStore::externGetNodeData(const storeId &id, bool &correctId) {
	QMutexLocker locker(&nodeStoreLock);
	return getNodeData(id, correctId);
}

//internal function (no lock)
BaseNode NodeStore::getBaseNode(const storeId &id, bool &correctId) {
	map<storeId, BaseNode>::iterator it = nodeMap.find(id);
	if(it!=nodeMap.end()) {
		correctId=true;
		if(it->second.validPointer()!=true) {
			cout << "NodeStore::getBaseNode : Invalid pointer of a BaseNode in getBaseNode id #" << id << " " << endl;
			correctId = false;
		}
		return it->second;
	} else {
		correctId=false;
		if(warningEnabled) {
			cout << "NodeStore::getBaseNode : Trying to get the BaseNode of an incorrect id #" << id << " " << endl;
		}
		BaseNode a;
		return a;
	}
}

//internal function (no lock)
//get the nodeData corresponding to a given id
NodeData NodeStore::getNodeData(const storeId &id, bool &correctId) {
	map<storeId, NodeData>::iterator it = nodeDataMap.find(id);
	if(it!=nodeDataMap.end()) {
		correctId=true;
		return it->second;
	} else {
		correctId=false;
		if(warningEnabled) {
			cout << "NodeStore::getNodeData : Trying to get the NodeData of an incorrect id #" << id << " " << endl;
		}
		NodeData a;
		return a;
	}
}

//set or update the nodeData of a given id
void NodeStore::setNodeData(const storeId &id, NodeData d) {
	QMutexLocker locker(&nodeStoreLock);
	if(id<0) return; //usually -1, a node with an uninitialized value
	
	if(debugLevel==4) {
		cout << "(setNodeData) Id #" << id << " value " << d.studied_pos_index << " / " << d.children_number << endl;
	}
	nodeDataMap[id] = d;
}

//external function : call the internal function with a lock
list<storeId> NodeStore::externGetChildrenList(const storeId &id, bool &correctId) {
	QMutexLocker locker(&nodeStoreLock);
	return getChildrenList(id, correctId);
}

//internal function (no lock)
list<storeId> NodeStore::getChildrenList(const storeId &id, bool &correctId) {
	map< storeId, list<storeId> >::iterator it = childrenLink.find(id);
	if(it!=childrenLink.end()) {
		correctId=true;
		return it->second;
	} else {
		correctId=false;
		if(warningEnabled) {
			cout << "NodeStore::getChildrenList : Trying to get" << 
			        "the children list of an incorrect id #" << id << " " << endl;
		}
		list<storeId> a;
		return a;
	}
}

//internal function (no lock)
list<storeId> NodeStore::getParentList(const storeId &id, bool &correctId) {
	map< storeId, list<storeId> >::iterator it = parentLink.find(id);
	if(it!=parentLink.end()) {
		correctId=true;
		return it->second;
	} else {
		correctId=false;
		if(warningEnabled) {
			cout << "NodeStore::getParentList : Trying to get" <<
			        " the parent list of an incorrect id #" << id << " " << endl;
		}
		list<storeId> a;
		return a;
	}
}

list<BaseNode> NodeStore::getChildrenBaseNode(const storeId &id, bool &correctId) {
	QMutexLocker locker(&nodeStoreLock);
	list<BaseNode> result;
	map< storeId, list<storeId> >::iterator it = childrenLink.find(id);
	if(it!=childrenLink.end()) {
		correctId=true;
		list<storeId>::iterator it2;
		for(it2=it->second.begin() ; it2!=it->second.end() ; it2++) {
			BaseNode c = getBaseNode(*it2, correctId);
			result.push_back(c);
		}
	} else {
		correctId=false;
		cout << "NodeStore::getChildrenBaseNode : Trying to get the children BaseNode list of an incorrect id #" << id << " " << endl;
	}
	return result;
}

//internal function (no lock)
//return the index of the given children in the children list of the parent
int NodeStore::childrenIndex(const storeId &parentId, const storeId &childrenId) {
	bool correctId; //unused
	list<storeId> childrenIdList = getChildrenList(parentId , correctId);
	list<storeId>::iterator it;
	int i;
	for(it=childrenIdList.begin(), i=0; it!=childrenIdList.end(); it++, i++) {
		if(*it == childrenId) {
			return i;
		}
	}
	
	//error case
	return -1;
}

//internal function (no lock)
//compute the bcu of a given node (-1 if the value is not known)
//bcu = "best child unknown" = the number of children of the child with the least number of children
int NodeStore::computeBcu(const storeId &id) {
	bool correctId;
	list<storeId> childrenIdList = getChildrenList(id , correctId);
	list<storeId>::iterator it;
	int bcu=-1;
	for(it=childrenIdList.begin(); it!=childrenIdList.end(); it++) {
		NodeData a = getNodeData(*it, correctId);
		if(correctId) {
			//don't take into account already killed children
			if(a.unknown_number>0 && !a.knownState) {
				//update the bcu if it has not been updated yet or if the new unknown number is smaller
				if(bcu==-1 || a.unknown_number<bcu) bcu = a.unknown_number;
			}
		}
	}
	return bcu;
}

//------------NodeStore computing branch---------------------

//add a new id at the end of the computing branch
void NodeStore::compBranchAdd(const storeId &id) {
	Trace::create_level();
	
	QMutexLocker locker(&nodeStoreLock);
	computingBranch.push_back(id);
	lastComputedId = -1;
	if(debugLevel==4) {
		cout << "(compBranchAdd) ";
		printList(computingBranch);
	}
}

//delete the last id of the computing branch
void NodeStore::compBranchDel() {
	Trace::delete_level();
	
	QMutexLocker locker(&nodeStoreLock);
	if(!computingBranch.empty()) {
		lastComputedId = computingBranch.back();
		computingBranch.pop_back();
	}
	if(debugLevel==4) {
		cout << "(compBranchDel) ";
		printList(computingBranch);
	}
}

//update the last id of the computing branch
void NodeStore::compBranchUpd(const storeId &id) {
	QMutexLocker locker(&nodeStoreLock);
	list<storeId>::reverse_iterator it = computingBranch.rbegin();
	if(it!=computingBranch.rend()) {
		*it=id;
		lastComputedId = -1;
	} else {
		cout << "Trying to update an empty NodeStore::computingBranch" << endl;
	}
	if(debugLevel==4) {
		cout << "(compBranchUpd) ";
		printList(computingBranch);
	}
}

storeId NodeStore::compBranchLast() {
	QMutexLocker locker(&nodeStoreLock);
	if(!computingBranch.empty()) {
		return computingBranch.back();
	} else {
		cout << "Trying to get the last id of an empty computing branch in NodeStore::compBranchLast" << endl;
		return 0;
	}
}

bool NodeStore::compBranchIsEmpty() {
	QMutexLocker locker(&nodeStoreLock);
	return computingBranch.empty();
}

int NodeStore::compBranchSize() {
	QMutexLocker locker(&nodeStoreLock);
	int s = computingBranch.size();
	return s;
}

//internal function (no lock)
list<storeId> NodeStore::getComputingBranch() {
	return computingBranch;
}

//--------------------------------Trace ---------------------------------------
//-----------------------------------------------------------------------------

//return a string with global informations about the nodestore state
QString NodeStore::stateString() {
	QMutexLocker locker(&nodeStoreLock);
	QString result("    NodeStore State :\n");
	if(useTraversal) {
		result += QString("    TR nodes   : %1\n").arg(TRnodesCount);
	} else {
		result += QString("\n");
	}
	result += QString("    DF nodes   : %1\n").arg(DFnodesCount);
	result += QString("    Iterations : %1").arg(mainloopCount);
	return result;
}

//get all the information needed for the main trace with only one function
list<NodeTrace> NodeStore::getMainTrace() {
	QMutexLocker locker(&nodeStoreLock);
	list<NodeTrace> completeResult;
	
	list<storeId> listS = NodeStore::getComputingBranch();
	list<storeId>::iterator it, previousIt;
	
	//main loop on the list of ids in the computing Branch
	for(it=listS.begin(); it!=listS.end(); it++) {
		NodeTrace result;
		result.posColor = QString("transparent");
		result.traversal_info = QString("");
		result.is_alive = true;
		
		previousIt=it;
		if(previousIt!=listS.begin()) {
			previousIt--;
		}
		
		bool correctId;
		NodeData d = NodeStore::getNodeData(*previousIt, correctId);
		if(!correctId) return completeResult;
		
		NodeData d_current = NodeStore::getNodeData(*it, correctId);
		if(!correctId) return completeResult;
		
		BaseNode c = NodeStore::getBaseNode(*it, correctId);
		if(!correctId) return completeResult;
		
		//index
		if(it!=listS.begin()) {
			result.index  = QString("%1 ").arg( d.studied_pos_index + 1, 2, 10 );
			result.index += QString("/");
			result.index += QString("%1 ").arg(d.children_number, 2, 10 );
			result.alive =d.unknown_number; 
		} else {
			//particular case of the first node
			result.index = QString("%1 ").arg( 1, 2, 10 ) + QString("/") + QString("%1 ").arg( 1, 2, 10 );
			result.alive = 1;
		}
		
		//Tree Traversal
		//print only the data of traversal-type nodes, and the data of the root node
		if(useTraversal && (d_current.traversal_data.isTraversalNode() || it==listS.begin())) {
			result.traversal_info = d_current.traversal_data.getMainTrace();
		}
		
		//number of lives, nimber, position
		QStringList disp = c.displayStringList();
		result.nb_lives = disp.at(0);
		result.pos = disp.at(1);
		
		result.posColor = c.displayColor();	//color of the position
		
		//special color for a traversal node chosen by the user
		if(useTraversal && *it == userChosenId) result.posColor = QString("light green");
		
		//TODO : the value is probably incorrect
		MoveOrdering param = c.getMoveOrdering();
		if(param.checkValue >= 101) {
			//losing child in the previous computation
			result.posColor = QString("blue");
		}
		
		completeResult.push_back(result);
	}
	
	return completeResult;
}

//return the storeId of the node currently computed at a given level of depth
storeId NodeStore::getLevelStoreId(int level) {
	//find the requested (trace) level
	nodeStoreLock.lock();
	list<storeId> listS = NodeStore::getComputingBranch();
	nodeStoreLock.unlock();
	
	list<storeId>::iterator it = listS.begin();
	int i;
	for(i=0; i<level && it!=listS.end(); i++) {
		it++;
	}
	if(it==listS.end()) return 0; //the requested level doesn't exist, 0 by convention
	
	return *it;
}

//get all the information needed for the children trace with only one function
list<NodeTrace> NodeStore::getChildrenTrace(const storeId &id) {
	QMutexLocker locker(&nodeStoreLock);
	warningEnabled=false; //disable the warning messages if we try to access to nodes that don't exist
	list<NodeTrace> completeResult;
	//get the list of children for the requested id
	bool correctId;
	list<storeId> childrenId = NodeStore::getChildrenList(id, correctId);
	if(!correctId) { warningEnabled = true; return completeResult; }
	completeResult = traceFromList(childrenId);
	warningEnabled = true;
	return completeResult;
}

list<NodeTrace> NodeStore::getParentsTrace(const storeId &id) {
	QMutexLocker locker(&nodeStoreLock);
	warningEnabled=false; //disable the warning messages if we try to access to nodes that don't exist
	list<NodeTrace> completeResult;
	//get the list of parents for the requested id
	bool correctId;
	list<storeId> parentId = NodeStore::getParentList(id, correctId);
	if(!correctId) { warningEnabled = true; return completeResult; }
	completeResult = traceFromList(parentId);
	warningEnabled = true;
	return completeResult;
}

//get the information of only one node given in parameter
//we return a list to have compatibility with getParentsTrace and getChildrenTrace
list<NodeTrace> NodeStore::getOneNodeTrace(const storeId &id) {
	QMutexLocker locker(&nodeStoreLock);
	warningEnabled=false; //disable the warning messages if we try to access to nodes that don't exist
	list<storeId> currentId;
	currentId.push_back(id);
	list<NodeTrace> completeResult = traceFromList(currentId);
	warningEnabled = true;
	return completeResult;
}

//no lock, internal function
list<NodeTrace> NodeStore::traceFromList(const list<storeId> &list_id) {
	list<NodeTrace> completeResult;
	bool correctId;
	list<storeId> computingIds = NodeStore::getComputingBranch();
	list<storeId>::iterator computingIt;
	list<storeId>::const_iterator childIt;
	for(childIt=list_id.begin(); childIt!=list_id.end(); childIt++) {
		NodeTrace result;
		result.posColor = QString("transparent"); //background color for the position
		
		BaseNode c = NodeStore::getBaseNode(*childIt, correctId);
		if(!correctId) return completeResult;
		
		//storeid
		result.id = QString("%1").arg(*childIt, 1, 10 );
		
		//number of lives & position
		QStringList disp = c.displayStringList();
		result.nb_lives = disp.at(0);
		result.pos = disp.at(1);
		
		//Tree Traversal
		result.traversal_info = QString("");
		result.alive =  0;
		result.bcu =  0;
		result.is_alive = true;
		NodeData d2 = NodeStore::getNodeData(*childIt, correctId);
		if(correctId) {
			//number of still unknown children
			result.alive = d2.unknown_number;
			result.bcu = NodeStore::computeBcu(*childIt);
			
			if(useTraversal) result.traversal_info = d2.traversal_data.getSecondaryTrace();
			
			if(d2.knownState) result.is_alive = false; //already known child
			
			//look for this node in the computing branch : if we are currently computing it, background color is blue
			computingIt = find(computingIds.begin(), computingIds.end(), /*current id :*/ *childIt);
			if(computingIt!=computingIds.end()) result.posColor = QString("blue");
			
			//special case of a traversal node chosen by the user
			if(useTraversal && *childIt == userChosenId) result.posColor = QString("light green");
		}
		
		completeResult.push_back(result);
	}
	
	return completeResult;
}

//debug
//return a string with all the database sizes
string NodeStore::storeSize() {
	string result("Size ");
	result += QString("id:%1 ").arg(idMap.size()).toStdString();
	result += QString("node:%1 ").arg(nodeMap.size()).toStdString();
	result += QString("nodeData:%1 ").arg(nodeDataMap.size()).toStdString();
	result += QString("children:%1 ").arg(childrenLink.size()).toStdString();
	result += QString("parent:%1 ").arg(parentLink.size()).toStdString();
	return result;
}
